function [Bham] = BnBmain(s1, s2)
% Author: Yang Qi, Department of Biochemistry and Department of Computer
%         Science, Duke University
%
% main function to run the Branch and Bound algorithm
% the function returns the parameters of the best-fit Bingham distribution
% input: Saupe tensors of two alignments for each domain
%        in the form of [Sxx1,Syy1,Sxy1,Sxz1,Syz1;Sxx2,Syy2,Sxy2,Sxz2,Syz2]
%        s1 contains the Saupe tensors of the reference domain
%        s2 contains the Saupe tensors of the other domain
%        orthogonal tensors are highly recommended, but not required
% output: the parameters of the best-fit Bingham distribution(Ls,quaL,quaR)
% requires iBranch, branch, r3prune, rsprune, enum, ijk2qua

% Acknowledgements:
% the mhg function: Copyright (c) 2004 Plamen Koev
% Plamen Koev and Alan Edelman, The Efficient Evaluation of the 
% Hypergeometric Function of a Matrix Argument, Mathematics of Computation
% 75 (2006), 833-846.
% quaternions.txt is generated using the following sampling method
% Anna Yershova, Swati Jain, Steven M. LaValle and Julie C. Mitchell, 
% Generating Uniform Incremental Grids on SO(3) Using the Hopf Fibration,
% The International Journal of Robotics Research 29 (2010), 801812.

% when no input or incorret input, the SpA-N data is used
if nargin ~= 2
    disp('invalid input! SpA-N data is used as default')
    % Orthogonal S tensors of both Z and C domains
    % Saupe tensors are in the form of {Sxx,Syy,Sxy,Sxz,Syz}
    s1 = [-8.6551,  14.9983, -20.564, -0.41839, 0.36701;
        5.04792, -2.43413, -1.5583,  2.56783, 2.28739]';
    
    s2 = [-0.521156,  1.57063,  0.771357,  0.0267095,  0.52418;
        0.357427, -0.54729,  0.036935, -0.0785261, -0.45442]';
end

% define the search range and search resolution
% intRegn = [-4,4;-4,4;-4,4]; % total search region on R^{3}
intRegn = [-4,-3.5;-1,-0.5;1.5,2]
Rnlvl = 1; % pruning depth on R^{3}; resolution of 0.25
Snlvl = 1; % pruning depth on S^{3}; resolution of 0.13 rad

% *************************************************************************
% 1st step pruning on R^{3}

% The initial branch
[regns, iregns] = iBranch(intRegn);

% branching and prunning on both legal and illegal regions
for nBranch = 1:(Rnlvl-1)
    % pruning step
    regns = r3prune(regns, s1, s2, false);
    
    % branching step
    % branch on legal regions
    regns1 = zeros(3,2,0);
    for i = 1:size(regns,3)
        subRegns = branch(regns(:,:,i));
        regns1 = cat(3,regns1,subRegns);
    end
    % branch on illegal regions
    regns2 = zeros(3,2,0);
    iregns2 = zeros(3,2,0);
    for j = 1:size(iregns,3)
        [subRegns,subiRegns] = iBranch(iregns(:,:,j));
        regns2 = cat(3,regns2,subRegns);
        iregns2 = cat(3,iregns2,subiRegns);
    end
    
    regns = cat(3,regns1,regns2);
    iregns = iregns2;
end
% pruning step
regns = r3prune(regns, s1, s2, true);

disp('The first pruning step completed! The pruning efficiency of the first step is:')
disp(1-size(regns,1)/(8^Rnlvl))

% *************************************************************************
% second pruning step on R^{3} x S^{3}
for lvl = 0:Snlvl
    regns = rsprune(regns,s1,s2,lvl);
end

disp('The second pruning step completed! The pruning efficiency of the first step is:')
disp(1-size(regns,1)/(8^Rnlvl*6912*8^Snlvl))

% *************************************************************************
% enumeration step
regns = enum(regns,s1,s2,Snlvl);
disp('The enumeration step completed.')

% *************************************************************************
% summarize and save the pruning result

% percent of the pruned volume
Nregns = size(regns,1);
percent = 1 - Nregns/(8^Rnlvl*6912*8^Snlvl);

% find the minimum point
objVals = zeros(1,Nregns);
for ii = 1:Nregns
    objVals(1,ii) = regns{ii,3};
end

[~,idx] = min(objVals);

if Nregns ~= 0
    fRegn = regns(idx,:);
    Ls = sum(fRegn{1}(:,1:end-1),2)'/2;
    Rqua = ijk2qua(fRegn{1}(1,end),fRegn{1}(2,end),fRegn{1}(3,end),Snlvl);
    Lqua = fRegn{2};
    Bham.Ls = Ls;
    Bham.quaL = Rqua.*[-1,1,1,1];
    Bham.quaR = Lqua.*[-1,1,1,1];
    disp('Paramters of the best-fit Bignham distribution(Ls, quaL, quaR):')
    disp(Bham)
end

disp('Number of enumerated regions:')
disp(Nregns)
disp('Pruning effciency:')
disp(percent)

save('result.mat','regns','objVals');
save('report.mat','Bham','Nregns','percent');
disp('result saved!')
end
