function [Qave] = calQave(L)
% Author: Yang Qi, Department of Biochemistry and Department of Computer
%         Science, Duke University
%
% calculate the expectation of Q matrix over SO(3)
% the integration is calculated by approximating hypergeometric functions
% input: L, a row vector containing the concentration parameters
% output: Qave, a 5x5 matrix

% calculating the expectation equals calculating Q matrix's second-order derivatives
delta = 0.00001; % step size for doing numerical differentiation
n = 1000; % defines the accuracy of the hypergeometric approximation

% calculating hessians
r12 = PAniso(1,2,L,delta,n);
r34 = PAniso(3,4,L,delta,n);
r13 = PAniso(1,3,L,delta,n);
r24 = PAniso(2,4,L,delta,n);
r14 = PAniso(1,4,L,delta,n);
r23 = PAniso(2,3,L,delta,n);

% assemble the hessians into the Q matrix
m11 = 1/3*(-r12-r34+3*r13+3*r24+r14+r23);
m12 = 1/3*(2*r13+2*r24-2*r14-2*r23);
m21 = 1/3*(2*r12+2*r34-2*r14-2*r23);
m22 = 1/3*(3*r12+3*r34-r13-r24+r14+r23);
m33 = r14-r23;
m44 = r13-r24;
m55 = r12-r34;

Qave = [m11,m12,0,0,0;m21,m22,0,0,0;0,0,m33,0,0;0,0,0,m44,0;0,0,0,0,m55];
end

function [r] = PAniso(h1,h2,L,delta,n)

hesD1 = HesD(h1,L,delta,n);
hesD2 = HesD(h2,L,delta,n);
hesC12 = HesC(h1,h2,L,delta,n);

r = hesD1+hesD2-6*hesC12;

end

function [hesD]=HesD(h1,L,delta,n)
% calculate diagonal hessians

hs = [1,2,3,4];
hs = hs(hs~=h1);

a = L(h1);
b = L(hs(1));
c = L(hs(2));
d = L(hs(3));

% compute the truncated Hypergeometric functions of one matrix argument
ya1 = mhg(n,2,[0.5],[2],[a,b,c,d]);
ya2 = mhg(n,2,[0.5],[2],[a+delta,b,c,d]);
ya3 = mhg(n,2,[0.5],[2],[a+2*delta,b,c,d]);
hesD = (ya3+ya1-2*ya2)/(ya1*delta^2);

end

function [hesC]=HesC(h1, h2, L,delta,n)
% calculate off-diagonal hessians

hs = [1,2,3,4];
hs = hs(hs~=h1);
hs = hs(hs~=h2);

a = L(h1);
b = L(h2);
c = L(hs(1));
d = L(hs(2));

% compute the truncated Hypergeometric functions of one matrix argument
yab1 = mhg(n,2,[0.5],[2],[a,b,c,d]);
yab2 = mhg(n,2,[0.5],[2],[a,b+delta,c,d]);
yab3 = mhg(n,2,[0.5],[2],[a+delta,b,c,d]);
yab4 = mhg(n,2,[0.5],[2],[a+delta,b+delta,c,d]);
hesC=((yab4-yab3)-(yab2-yab1))/(yab1*(delta)^2);

end