function [pRegns] = enum(regns,s1,s2,Snlvl)
% Author: Yang Qi, Department of Biochemistry and Department of Computer
%         Science, Duke University
%
% enumerate all remaining regions and return solutions
% solutions with objective function values under ep are returned
% input: remaining regns from rsprune (regns), saupe tensor arrays (s1, s2),
%        level of branching on S^{3} (Snlvl)
% output: cell array pRegns, each row contains a 3x3 matrix, quaL and
%         objective function value(objVal)
%         the matrix contains the range of Ls and the index of quaR
%         the quaL and quaR here are NOT the output Bham.quaL and Bham.quaR
%         in the main function!
% requires ijk2qua, objFunc, calQm, v2m
pRegns = cell(0,3);
ep = 1; % objective function cutoff
Nregns = size(regns,1);

p = 1;
for m = 1:Nregns
    regn = regns{m,1};
    mQave = regns{m,6};
    
    i = regn(1,3);
    j = regn(2,3);
    k = regn(3,3);
    
    quaR = ijk2qua(i,j,k,Snlvl);
    
    [~,RL1,RL2] = calErr(s1,s2,mQave,quaR);
    
    quaLs = twoR2quas(RL1,RL2);
    NquaL = size(quaLs,1);
    objVals = zeros(1,NquaL);
    for n = 1:NquaL
        objVals(n) = objFunc(s1,s2,mQave,quaLs(n,:),quaR);
    end
    [objVal,idx] = min(objVals);
    quaL = quaLs(idx,:);
    
    if objVal<ep
        pRegns{p,1} = regn;
        pRegns{p,2} = quaL;
        pRegns{p,3} = objVal;
        p = p+1;
    end
end

end

function [error,R1,R2] = calErr(s1,s2,Qave,quaR)
    QR = calQm(quaR);
    s1q = Qave*QR*s1;

    vqs = calVs(s1q);
    v2s = calVs(s2);
    
    RLs = RotV(vqs,v2s);
    
    [error,R1,R2] = calDist(RLs);
end

function [vs] = calVs(s)
    Ns = size(s,2);
    vs = zeros(3,3,Ns);

    for i = 1:Ns
        sm = v2m(s(:,i));
        [v,~] = eigs(sm);
        if det(v)<0
            v = -v;
        end
        
        vs(:,:,i) = v';
    end
end

function [Rs] = RotV(vqs,v2s)
    Ns = size(vqs,3);
    Rs = zeros(3,3,4,Ns);
    Mflip = zeros(3,3,4);
    Mflip(:,:,1) = [1,0,0;0,1,0;0,0,1];
    Mflip(:,:,2) = [-1,0,0;0,-1,0;0,0,1];
    Mflip(:,:,3) = [-1,0,0;0,1,0;0,0,-1];
    Mflip(:,:,4) = [1,0,0;0,-1,0;0,0,-1];
    
    % v2s = vqs*R
    for i = 1:Ns
        for j = 1:4
            Rs(:,:,j,i) = vqs(:,:,i)'*Mflip(:,:,j)*v2s(:,:,i);
        end
    end
end

function [dist,R1,R2] = calDist(Rs)
    Ns = size(Rs,4);
    distM = zeros(Ns,Ns);
    idxs = cell(Ns,Ns);
    
    for i = 1:Ns-1
        for j = i+1:Ns
            dists = zeros(4,4);
            for k = 1:4
                for l = 1:4
                    dists(k,l) = sumsqr(Rs(:,:,k,i)-Rs(:,:,l,j));
                end
            end
            [distM(i,j),ind1] = min(dists(:));
            [idxk,idxl] = ind2sub(size(dists),ind1);
            idxs{i,j} = [idxk,idxl];
        end
    end
    
    [dist,ind2] = max(distM(:));
    [idxi,idxj] = ind2sub(size(distM),ind2);
    idxk = idxs{idxi,idxj}(1);
    idxl = idxs{idxi,idxj}(2);
    R1 = Rs(:,:,idxk,idxi);
    R2 = Rs(:,:,idxl,idxj);
end

function [quaLs] = twoR2quas(RL1,RL2)
    quaL1 = R2qua(RL1);
    quaL2 = R2qua(RL2);
    if quaL1*quaL2' > 0.95
        quaLs = 0.5*quaL1+0.5*quaL2;
        quaLs = quaLs/norm(quaLs);
    elseif quaL1*quaL2' < -0.95
        quaLs = 0.5*quaL1-0.5*quaL2;
        quaLs = quaLs/norm(quaLs);
    else
        if quaL1*quaL2' > 0
            a3 = 0.5*quaL1+0.5*quaL2;
        else
            a3 = 0.5*quaL1-0.5*quaL2;
        end
        a3 = a3/norm(a3);
        
        quaLs = zeros(9,4);
        [Q,~] = qr([quaL1',quaL2',eye(4)]);
        a1 = Q(:,3)';
        a2 = Q(:,4)';
        quaLs(1,:) = a1;
        quaLs(2,:) = a2;
        quaLs(3,:) = a3;
        quaLs(4,:) = a1+a2/norm(a1+a2);
        quaLs(5,:) = a1-a2/norm(a1-a2);
        quaLs(6,:) = a1+a3/norm(a1+a3);
        quaLs(7,:) = a2+a3/norm(a2+a3);
        quaLs(8,:) = a3-a1/norm(a3-a1);
        quaLs(9,:) = a3-a2/norm(a3-a2);
    end
end

function [qua] = R2qua(R)
    if det(R)<0
        R = -R;
    end
    qua = zeros(1,4);
    qua(1) = 1/2*sqrt(1+R(1,1)+R(2,2)+R(3,3));
    qua(2) = 1/4*(R(3,2)-R(2,3))/qua(1);
    qua(3) = 1/4*(R(1,3)-R(3,1))/qua(1);
    qua(4) = 1/4*(R(2,1)-R(1,2))/qua(1);
    qua = qua/norm(qua);
end