function [pRegns] = r3prune(regns, s1, s2, boolQ)
% Author: Yang Qi, Department of Biochemistry and Department of Computer
%         Science, Duke University
%
% prune regions in R^3 (Ls)
% input: an 3x2xn array (regns), saupe tensor arrays (s1, s2), boolean to
%        specify output type
%        each element in the array is a 3x2 matrix containing the range of Ls
% output: an 3x2xn array (boolQ = false) or a cell array(boolQ = true)
%         each row in the cell array contains the range of Ls, uQave,
%         lQave, mQave
% requires calQave, calQm, calws, s2da, v2m

% remaining regions
if boolQ % choose whether to store uQave and lQave and mQave
    pRegns = cell(0,4);
else
    pRegns = zeros(3,2,0);
end

Nregns = size(regns,3);

if Nregns == 0
    return
end

ep = 0.25; % to counter numerical error
Da2s = s2da(s2);
Ns = size(s1,2);
dataQua = importdata('quaternions.txt'); % import quaternion list

p = 1;
for i = 1:Nregns
    regn = regns(:,:,i);
    [maxDas,minDas,criDas,uQave,lQave,mQave] = bounds(regn,s1,dataQua);
    inBound = true;
    for j = 1:Ns
        if (maxDas(j)+ep>Da2s(j) && criDas(j)-ep<Da2s(j)) || (-criDas(j)+ep>Da2s(j) && minDas(j)-ep<Da2s(j))
        else
            inBound = false;
        end
    end

    if inBound
        if boolQ
            pRegns{p,1} = regn;
            pRegns{p,2} = uQave;
            pRegns{p,3} = lQave;
            pRegns{p,4} = mQave;
        else
            pRegns(:,:,p) = regn;
        end
        p = p+1;
    end
end

end

function [maxDas,minDas,criDas,uQave,lQave,mQave] = bounds(regn,s1,dataQua)
    % biggest lamda1 -> least averaging -> upper bound
    uL = [-sum(regn(:,1)),regn(1,1),regn(2,1),regn(3,1)];
    uQave = calQave(uL);
    % smallest lamda1 -> most averaging -> lower bound
    lL = [-sum(regn(:,2)),regn(1,2),regn(2,2),regn(3,2)];
    lQave = calQave(lL);
    
    % calculate the Qave of the middle point in the region
    % prepare for the enumeration step
    mL = mean([uL;lL]);
    mQave = calQave(mL);
    
    nQua = size(dataQua,1);
    Ns = size(s1,2);
    s1m = v2m(s1);
    maxDas = zeros(1,Ns);
    minDas = zeros(1,Ns);
    criDas = zeros(1,Ns);
    for i = 1:Ns
        [u,w,v] = svd(s1m(:,:,i));
        for j = 1:3
            w(j,j) = u(:,j)'*v(:,j)*w(j,j);
        end
        s10 = [w(3,3),w(2,2),0,0,0]';
        
        if w(1,1)>0
            plusDa = true;
        else
            plusDa = false;
        end
        
        ws = calws(s10,[1,0,0,0],uQave);
        [~,idx] = max(abs(ws));
        maxDa0 = ws(idx);

        maxWs = zeros(1,nQua);
        criWs = zeros(1,nQua);
        
        for j = 1:nQua
            quaR = dataQua(j,:);
            uWs = calws(s10,quaR,uQave);
            lWs = calws(s10,quaR,lQave);
            
            [~,idx] = max(abs(uWs));
            maxWs(j) = uWs(idx);
            criWs(j) = 0.5*sum(abs(lWs));
        end
        maxDa = max(maxWs);
        minDa = min(maxWs);
        criDa = min(criWs);
        
        % exchange max and min if the directory component is negative
        if plusDa
            maxDas(i) = maxDa0;
            minDas(i) = minDa;
            criDas(i) = criDa;
        else
            maxDas(i) = maxDa;
            minDas(i) = maxDa0;
            criDas(i) = criDa;
        end
    end
end