function [pRegns] = rsprune(regns, s1, s2, lvl)
% Author: Yang Qi, Department of Biochemistry and Department of Computer
%         Science, Duke University
%
% prune regions in R^3 x S^3(q_{L})
% input: remaining regns from r3prune or rsprune (regns), saupe tensor
%        arrays (s1, s2), the current level of branching on S^{3} (lvl)
% output: cell array pRegns, each row contains a 3x3 matrix, a 2x2 matrix,
%         uQave, lQave, a local connection map(intermediate), mQave
%         the 3x3 matrix contains the range of Ls and the index of quaR
%         the 2x2 matrix contains the upper and lower bounds of the directory
%         component of each alignment
% requires s2da, calws

% remaining regions
pRegns = cell(0,6);
% pruned regions
qRegns = cell(0);

Nregns = size(regns,1);
if Nregns == 0
    return
end

Da2s = s2da(s2);
Ns = size(s1,2);

% grid resolution on S^3
ni = 12*2^lvl;
nj = 24*2^lvl;
nk = 24*2^lvl;

p = 1;
q = -1;
if lvl == 0
    for m = 1:Nregns
        regn = regns(m,:);
        % first-level prune
        [pRegns,p] = mprune(s1,Da2s,regn,pRegns,ni,nj,nk,Ns,p);
    end
else
    for m = 1:Nregns
        % other-level prune
        [pRegns,qRegns,regns,p,q] = cprune(s1,Da2s,regns,pRegns,qRegns,ni,nj,nk,Ns,lvl,p,q,m);
    end
    
    pRegns = neg2zero(pRegns,lvl);
end

end

function [pRegns,p] = mprune(s1,Da2s,regn,pRegns,ni,nj,nk,Ns,p)
    regnMap = zeros(ni,nj,nk);
    p0 = p;
    
    uMaxWs = zeros(ni,nj,nk,Ns);
    lMaxWs = zeros(ni,nj,nk,Ns);
    cLvl = 10;
    
    uQave = regn{2};
    lQave = regn{3};
    mQave = regn{4};
    
    for i = 1:ni
        for j = 1:nj
            for k = 1:nk
                theR = pi*(i-1)/ni;
                phiR = 2*pi*(j-1)/nj;
                psiR = 2*pi*(k-1)/nk-pi;
                quaR = [cos(theR/2)*cos(psiR/2),cos(theR/2)*sin(psiR/2),sin(theR/2)*cos(phiR+psiR/2),sin(theR/2)*sin(phiR+psiR/2)];
                
                % calculating the upper and lower bounds of abs(Da) for
                % each alignment
                for n = 1:Ns
                    uWs = calws(s1(:,n),quaR,uQave);
                    lWs = calws(s1(:,n),quaR,lQave);
                    uMaxWs(i,j,k,n) = max(abs(uWs));
                    lMaxWs(i,j,k,n) = max(abs(lWs));
                end
            end
        end
    end
    
    % calculate the range of abs(Da) in each alignment and error bounds
    uEp = zeros(1,Ns);
    lEp = zeros(1,Ns);
    for n = 1:Ns
        subuMaxWs = uMaxWs(:,:,:,n);
        uW1 = max(subuMaxWs(:));
        uW2 = min(subuMaxWs(:));
        uEp(n) = uW1-uW2; 
        sublMaxWs = lMaxWs(:,:,:,n);
        lW1 = max(sublMaxWs(:));
        lW2 = min(sublMaxWs(:));
        lEp(n) = lW1-lW2;
    end
    
    for i = 1:ni
        for j = 1:nj
            for k = 1:nk
                inBound = true;
                for n = 1:Ns
                    if abs(Da2s(n))<uMaxWs(i,j,k,n)+uEp(n)/(2*cLvl) && abs(Da2s(n))>lMaxWs(i,j,k,n)-lEp(n)/(2*cLvl)
                    else
                        inBound = false;
                    end
                end
                
                if inBound
                    pRegns{p,1}(:,1:2) = regn{1};
                    pRegns{p,1}(:,3) = [i,j,k];
                    pRegns{p,2} = [uEp;lEp];
                    pRegns{p,3} = uQave;
                    pRegns{p,4} = lQave;
                    pRegns{p,6} = mQave;
                    % the global connection map on S^3
                    % remaining points: index = p
                    % pruned points: index = 0
                    regnMap(i,j,k) = p;
                    p = p+1;
                end
            end
        end
    end
    
    % construct a local connection map for each point on S^3
    p1 = p-1;
    for ip = p0:p1
        i = pRegns{ip,1}(1,3);
        j = pRegns{ip,1}(2,3);
        k = pRegns{ip,1}(3,3);
        
        % a local connection map
        % index: (:,:,:,1) lvl: (:,:,:,2)
        adjs = zeros(3,3,3,2);
        for ii = 1:3
            for jj = 1:3
                for kk = 1:3
                    idxi = i+ii-2;
                    idxj = j+jj-2;
                    idxk = k+kk-2;
                    if idxi==0
                        idxi = ni;
                    elseif idxi==ni+1
                        idxi = 1;
                    end
                    if idxj==0
                        idxj = nj;
                    elseif idxj==nj+1
                        idxj = 1;
                    end
                    if idxk==0
                        idxk = nk;
                    elseif idxk==nk+1
                        idxk = 1;
                    end
                        
                    adjs(ii,jj,kk,1) = regnMap(idxi,idxj,idxk);
                end
            end
        end
        
        pRegns{ip,5} = adjs;
    end
    
end

function [pRegns,qRegns,regns,p,q] = cprune(s1,Da2s,regns,pRegns,qRegns,ni,nj,nk,Ns,lvl,p,q,m)
    regn = regns(m,:);
    i = regn{1}(1,3);
    j = regn{1}(2,3);
    k = regn{1}(3,3);
    uEp = regn{2}(1,:);
    lEp = regn{2}(2,:);
    uQave = regn{3};
    lQave = regn{4};
    % local connection map
    adjs = regn{5};
    mQave = regn{6};
    % large local connection map for new points generation
    % for main points: pruned: idx1 = 0; remaining: idx1 = pre_p;
    % for main points: unexplored: idx2 = 0; explored: idx2 = 1;
    % for new points: missing: idx1 = 0; pruned: idx1 = q(<0); remaining: idx1 = p;
    % for new points: missing: idx2 = 0; generated: idx2 = 1;
    adjMap = zeros(5,5,5,2);
    adjMap(:,:,:,2) = (lvl-1)*ones(5,5,5);
    
    if lvl == 1
        cLvl = 25;
    else
        cLvl = 50;
    end
    
    % update the large connection map from generated new points
    for ii = 1:3
        for jj = 1:3
            for kk = 1:3
                if adjs(ii,jj,kk,2) == lvl
                    % ip: current_p or current_q (only new points)
                    ip = adjs(ii,jj,kk,1);

                    if ip < 0
                        adjMap(ii:ii+2,jj:jj+2,kk:kk+2,:) = qRegns{-ip};
                    else
                        adjMap(ii:ii+2,jj:jj+2,kk:kk+2,:) = pRegns{ip,5};
                    end
                end
            end
        end
    end
    
    % generate new points
    for ii = 1:3
        for jj = 1:3
            for kk = 1:3
                if adjs(ii,jj,kk,2) == lvl-1
                    quaR = iijjkk2qua(i,j,k,ii,jj,kk,ni,nj,nk);
                    
                    uMaxWs = zeros(1,Ns);
                    lMaxWs = zeros(1,Ns);
                    
                    for n = 1:Ns
                        uWs = calws(s1(:,n),quaR,uQave);
                        lWs = calws(s1(:,n),quaR,lQave);
                        uMaxWs(n) = max(abs(uWs));
                        lMaxWs(n) = max(abs(lWs));
                    end
                    
                    inBound = true;
                    
                    for n = 1:Ns
                        if abs(Da2s(n))<uMaxWs(n)+uEp(n)/(2*cLvl) && abs(Da2s(n))>lMaxWs(n)-lEp(n)/(2*cLvl)
                        else
                            inBound = false;
                        end
                    end
                    
                    % update the large connection map
                    adjMap(ii+1,jj+1,kk+1,2) = lvl;
                    adjMap(2*ii-1,2*jj-1,2*kk-1,1) = adjs(ii,jj,kk,1);
                    % generate new points
                    if inBound
                        pRegns{p,1}(:,1:2) = regn{1}(:,1:2);
                        pRegns{p,1}(:,3) = [2*i+ii-3,2*j+jj-3,2*k+kk-3];
                        pRegns{p,2} = [uEp;lEp];
                        pRegns{p,3} = uQave;
                        pRegns{p,4} = lQave;
                        pRegns{p,6} = mQave;
                        adjMap(ii+1,jj+1,kk+1,1) = p;
                        p = p+1;
                    else
                        qRegns{-q} = zeros(3,3,3,2);
                        adjMap(ii+1,jj+1,kk+1,1) = q;
                        q = q-1;
                    end
                end
            end
        end
    end
    
    % generate local connection maps for the new points
    for ii = 1:5
        for jj = 1:5
            for kk = 1:5
                if adjMap(ii,jj,kk,2)==lvl
                    % ip: current_p or current_q (only new points)
                    ip = adjMap(ii,jj,kk,1);
                    if ip < 0
                        %qRegns{-ip} = adjMap(ii:ii+2,jj:jj+2,kk:kk+2,:);
                        qRegns{-ip}(max(1,3-ii):min(3,7-ii),max(1,3-jj):min(3,7-jj),max(1,3-kk):min(3,7-kk),:) = adjMap(max(1,ii-1):min(5,ii+1),max(1,jj-1):min(5,jj+1),max(1,kk-1):min(5,kk+1),:);
                    else
                        pRegns{ip,5}(max(1,3-ii):min(3,7-ii),max(1,3-jj):min(3,7-jj),max(1,3-kk):min(3,7-kk),:) = adjMap(max(1,ii-1):min(5,ii+1),max(1,jj-1):min(5,jj+1),max(1,kk-1):min(5,kk+1),:);
                    end
                end
            end
        end
    end
    
    % update local connection maps of adjacent main points
    for ii = 1:3
        for jj = 1:3
            for kk = 1:3
                if adjs(ii,jj,kk,2)==lvl-1 && adjs(ii,jj,kk,1)~=0
                    % ip: pre_p (only main points)
                    ip = adjs(ii,jj,kk,1);
                    regns{ip,5}(max(1,5-2*ii):min(3,7-2*ii),max(1,5-2*jj):min(3,7-2*jj),max(1,5-2*kk):min(3,7-2*kk),:) = adjMap(max(2,2*ii-2):min(4,2*ii),max(2,2*jj-2):min(4,2*jj),max(2,2*kk-2):min(4,2*kk),:);
                end
            end
        end
    end
end

function [pRegns] = neg2zero(pRegns,lvl)
    for m = 1:size(pRegns,1)
        adjs1 = pRegns{m,5}(:,:,:,1);
        adjs2 = pRegns{m,5}(:,:,:,2);
        for i = 1:27
            if adjs1(i)<0
                adjs1(i) = 0;
            end
            if adjs2(i)==lvl-1
                adjs2(i) = lvl;
            end
        end
        pRegns{m,5}(:,:,:,1) = adjs1;
        pRegns{m,5}(:,:,:,2) = adjs2;
    end
end

function [qua] = iijjkk2qua(i,j,k,ii,jj,kk,ni,nj,nk)
    theR = 2*pi*(i-1)/ni+(ii-2)*pi/ni;
    if i==1 && ii==1
        theR = theR + pi;
    end
    phiR = 4*pi*(j-1)/nj+(jj-2)*2*pi/nj;
    psiR = 4*pi*(k-1)/nk-pi+(kk-2)*2*pi/nk;
    if k==1 && kk==1
        psiR = psiR + 2*pi;
    end
    qua = [cos(theR/2)*cos(psiR/2),cos(theR/2)*sin(psiR/2),sin(theR/2)*cos(phiR+psiR/2),sin(theR/2)*sin(phiR+psiR/2)];
end